"""
Graceful failure for missing optional dependencies.
"""

import functools
import warnings


class ImportAlarmError(ImportError):
    """To be raised in addition to warning under test conditions"""


class ImportAlarm:
    """
    This class allows you to fail gracefully when some object has optional dependencies
    and the user does not have those dependencies installed.

    Example:

    >>> try:
    ...     from mystery_package import Enigma, Puzzle, Conundrum
    ...     import_alarm = ImportAlarm()
    >>> except ImportError:
    >>>     import_alarm = ImportAlarm(
    ...         "MysteryJob relies on mystery_package, but this was unavailable. Please ensure your python environment "
    ...         "has access to mystery_package, e.g. with `conda install -c conda-forge mystery_package`"
    ...     )
    ...
    >>> class MysteryJob:
    ...     @import_alarm
    ...     def __init__(self, project, job_name)
    ...         super().__init__()
    ...         self.riddles = [Enigma(), Puzzle(), Conundrum()]

    This class is also a context manager that can be used as a short-cut, like this:

    >>> with ImportAlarm(
    ...     "MysteryJob relies on mystery_package, but this was unavailable."
    ... ) as import_alarm:
    ...     import mystery_package

    If you do not use `import_alarm` as a decorator, but only to get a consistent
    warning message, call :meth:`.warn_if_failed()` after the with statement.

    >>> import_alarm.warn_if_failed()
    """

    def __init__(self, message=None, _fail_on_warning: bool = False):
        """
        Initialize message value.

        Args:
            message (str): What to say alongside your ImportError when the decorated
            function is called. (Default is None, which says nothing and raises no
            error.)
        """
        self.message = message
        # Catching warnings in tests can be janky, so instead open a flag for failing
        # instead.
        self._fail_on_warning = _fail_on_warning

    def __call__(self, func):
        return self.wrapper(func)

    def wrapper(self, function):
        @functools.wraps(function)
        def decorator(*args, **kwargs):
            self.warn_if_failed()
            return function(*args, **kwargs)

        return decorator

    def warn_if_failed(self):
        """
        Print warning message if import has failed.  In case you are not using
        :class:`ImportAlarm` as a decorator you can call this method manually to
        trigger the warning.
        """
        if self.message is not None:
            warnings.warn(self.message, category=ImportWarning)
            if self._fail_on_warning:
                raise ImportAlarmError(self.message)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None and exc_value is None and traceback is None:
            # import successful, so silence our warning
            self.message = None
            return
        if issubclass(exc_type, ImportError):
            # import broken; retain message, but suppress error
            return True
        else:
            # unrelated error during import, re-raise
            return False



"""
A utility  class for deprecating code.
"""

from copy import copy
import functools
import types
import warnings


class Deprecator:
    """
    Decorator class to mark functions and methods as deprecated with a uniform
    warning message at the time the function is called.  The message has the
    form

        {function_name} is deprecated: {message}. It is not guaranteed to be in
        service after {version}.

    unless `pending=True` was given.  Then the message will be

        {function_name} will be deprecated in a future version: {message}.

    If message and version are not initialized or given during the decorating
    call the respective parts are left out from the message.

    >>> deprecate = Deprecator()
    >>> @deprecate
    ... def foo(a, b):
    ...     pass
    >>> foo(1, 2)
    DeprecationWarning: __main__.foo is deprecated

    >>> @deprecate("use bar() instead")
    ... def foo(a, b):
    ...     pass
    >>> foo(1, 2)
    DeprecationWarning: __main__.foo is deprecated: use bar instead

    >>> @deprecate("use bar() instead", version="0.4.0")
    ... def foo(a, b):
    ...     pass
    >>> foo(1, 2)
    DeprecationWarning: __main__.foo is deprecated: use bar instead.  It is not
    guaranteed to be in service in vers. 0.4.0

    >>> deprecate = Deprecator(message="I say no!", version="0.5.0")
    >>> @deprecate
    ... def foo(a, b):
    ...     pass
    >>> foo(1, 2)
    DeprecationWarning: __main__.foo is deprecated: I say no!  It is not
    guaranteed to be in service in vers. 0.5.0

    Alternatively the decorator can also be called with `arguments` set to a dictionary
    mapping names of keyword arguments to deprecation messages.  In this case the
    warning will only be emitted when the decorated function is called with arguments
    in that dictionary.

    >>> deprecate = Deprecator()
    >>> @deprecate(arguments={"bar": "use baz instead."})
    ... def foo(bar=None, baz=None):
    ...     pass
    >>> foo(baz=True)
    >>> foo(bar=True)
    DeprecationWarning: __main__.foo(bar=True) is deprecated: use baz instead.

    As a short-cut, it is also possible to pass the values in the arguments dict
    directly as keyword arguments to the decorator.

    >>> @deprecate(bar="use baz instead.")
    ... def foo(bar=None, baz=None):
    ...     pass
    >>> foo(baz=True)
    >>> foo(bar=True)
    DeprecationWarning: __main__.foo(bar=True)  is deprecated: use baz instead.
    """

    def __init__(self, message=None, version=None, pending=False):
        """
        Initialize default values for deprecation message and version.

        Args:
            message (str): default deprecation message
            version (str): default version after which the function might be removed
            pending (bool): only warn about future deprecation, warning category will
            be PendingDeprecationWarning instead of DeprecationWarning
        """
        self.message = message
        self.version = version
        self.category = PendingDeprecationWarning if pending else DeprecationWarning

    def __copy__(self):
        cp = type(self)(message=self.message, version=self.version)
        cp.category = self.category
        return cp

    def __call__(self, message=None, version=None, arguments=None, **kwargs):
        depr = copy(self)
        if isinstance(message, types.FunctionType):
            return depr.__deprecate_function(message)
        else:
            depr.message = message
            depr.version = version
            depr.arguments = arguments if arguments is not None else {}
            depr.arguments.update(kwargs)
            return depr.wrap

    def _build_message(self):
        if self.category == PendingDeprecationWarning:
            message_format = "{} will be deprecated"
        else:
            message_format = "{} is deprecated"

        if self.message is not None:
            message_format += ": {}.".format(self.message)
        else:
            message_format += "."

        if self.version is not None:
            message_format += (
                " It is not guaranteed to be in service in vers. {}".format(
                    self.version
                )
            )

        return message_format

    def __deprecate_function(self, function):
        message = self._build_message().format(
            "{}.{}".format(function.__module__, function.__name__)
        )

        @functools.wraps(function)
        def decorated(*args, **kwargs):
            warnings.warn(message, category=self.category, stacklevel=2)
            return function(*args, **kwargs)

        return decorated

    def __deprecate_argument(self, function):
        message_format = self._build_message()

        @functools.wraps(function)
        def decorated(*args, **kwargs):
            for kw in kwargs:
                if kw in self.arguments:
                    warnings.warn(
                        message_format.format(
                            "{}.{}({}={})".format(
                                function.__module__, function.__name__, kw, kwargs[kw]
                            )
                        ),
                        category=self.category,
                        stacklevel=2,
                    )
            return function(*args, **kwargs)

        return decorated

    def wrap(self, function):
        """
        Wrap the given function to emit a DeprecationWarning at call time.  The warning
        message is constructed from the given message and version.  If
        :attr:`.arguments` is set then the warning is only emitted, when the decorated
        function is called with keyword arguments found in that dictionary.

        Args:
            function (callable): function to mark as deprecated

        Return:
            function: raises DeprecationWarning when given function is called
        """
        if not self.arguments:
            return self.__deprecate_function(function)
        else:
            return self.__deprecate_argument(function)


deprecate = Deprecator()
deprecate_soon = Deprecator(pending=True)



class DotDict(dict):
    def __getattr__(self, item):
        try:
            return self.__getitem__(item)
        except KeyError:
            raise AttributeError(
                f"{self.__class__.__name__} object has no attribute '{item}'"
            )

    def __setattr__(self, key, value):
        self[key] = value

    def __dir__(self):
        return set(super().__dir__() + list(self.keys()))

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        for k, v in state.items():
            self.__dict__[k] = v

    def to_list(self):
        """A list of values (order not guaranteed)"""
        return list(self.values())



from __future__ import annotations
from pathlib import Path
import shutil


def delete_files_and_directories_recursively(path):
    if not path.exists():
        return
    for item in path.rglob("*"):
        if item.is_file():
            item.unlink()
        else:
            delete_files_and_directories_recursively(item)
    path.rmdir()


def categorize_folder_items(folder_path):
    types = [
        "dir",
        "file",
        "mount",
        "symlink",
        "block_device",
        "char_device",
        "fifo",
        "socket",
    ]
    results = {t: [] for t in types}

    for item in folder_path.iterdir():
        for tt in types:
            try:
                if getattr(item, f"is_{tt}")():
                    results[tt].append(str(item))
            except NotImplementedError:
                pass
    return results


def _resolve_directory_and_path(
    file_name: str,
    directory: DirectoryObject | str | None = None,
    default_directory: str = ".",
):
    """
    Internal routine to separate the file name and the directory in case
    file name is given in absolute path etc.
    """
    path = Path(file_name)
    file_name = path.name
    if path.is_absolute():
        if directory is not None:
            raise ValueError(
                "You cannot set `directory` when `file_name` is an absolute path"
            )
        # If absolute path, take that of new_file_name regardless of the
        # name of directory
        directory = str(path.parent)
    else:
        if directory is None:
            # If directory is not given, take default directory
            directory = default_directory
        else:
            # If the directory is given, use it as the main path and append
            # additional path if given in new_file_name
            if isinstance(directory, DirectoryObject):
                directory = directory.path
        directory = directory / path.parent
    if not isinstance(directory, DirectoryObject):
        directory = DirectoryObject(directory)
    return file_name, directory


class DirectoryObject:
    def __init__(self, directory: str | Path | DirectoryObject):
        if isinstance(directory, str):
            self.path = Path(directory)
        elif isinstance(directory, Path):
            self.path = directory
        elif isinstance(directory, DirectoryObject):
            self.path = directory.path
        self.create()

    def create(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def delete(self, only_if_empty: bool = False):
        if self.is_empty() or not only_if_empty:
            delete_files_and_directories_recursively(self.path)

    def list_content(self):
        return categorize_folder_items(self.path)

    def __len__(self):
        return sum([len(cc) for cc in self.list_content().values()])

    def __repr__(self):
        return f"DirectoryObject(directory='{self.path}')\n{self.list_content()}"

    def get_path(self, file_name):
        return self.path / file_name

    def file_exists(self, file_name):
        return self.get_path(file_name).is_file()

    def write(self, file_name, content, mode="w"):
        with self.get_path(file_name).open(mode=mode) as f:
            f.write(content)

    def create_subdirectory(self, path):
        return DirectoryObject(self.path / path)

    def create_file(self, file_name):
        return FileObject(file_name, self)

    def is_empty(self) -> bool:
        return len(self) == 0

    def remove_files(self, *files: str):
        for file in files:
            path = self.get_path(file)
            if path.is_file():
                path.unlink()


class NoDestinationError(ValueError):
    """A custom error for when neither a new file name nor new location are provided"""


class FileObject:
    def __init__(self, file_name: str, directory: DirectoryObject = None):
        self._file_name, self.directory = _resolve_directory_and_path(
            file_name=file_name, directory=directory, default_directory="."
        )

    @property
    def file_name(self):
        return self._file_name

    @property
    def path(self):
        return self.directory.path / Path(self._file_name)

    def write(self, content, mode="x"):
        self.directory.write(file_name=self.file_name, content=content, mode=mode)

    def read(self, mode="r"):
        with open(self.path, mode=mode) as f:
            return f.read()

    def is_file(self):
        return self.directory.file_exists(self.file_name)

    def delete(self):
        self.path.unlink()

    def __str__(self):
        return str(self.path.absolute())

    def _resolve_directory_and_path(
        self,
        file_name: str,
        directory: DirectoryObject | str | None = None,
        default_directory: str = ".",
    ):
        """
        Internal routine to separate the file name and the directory in case
        file name is given in absolute path etc.
        """
        path = Path(file_name)
        file_name = path.name
        if path.is_absolute():
            # If absolute path, take that of new_file_name regardless of the
            # name of directory
            directory = str(path.parent)
        else:
            if directory is None:
                # If directory is not given, take default directory
                directory = default_directory
            else:
                # If the directory is given, use it as the main path and append
                # additional path if given in new_file_name
                if isinstance(directory, DirectoryObject):
                    directory = directory.path
            directory = directory / path.parent
        if not isinstance(directory, DirectoryObject):
            directory = DirectoryObject(directory)
        return file_name, directory

    def copy(
        self,
        new_file_name: str | None = None,
        directory: DirectoryObject | str | None = None,
    ):
        """
        Copy an existing file to a new location.
        Args:
            new_file_name (str): New file name. You can also set
                an absolute path (in which case `directory` will be ignored)
            directory (DirectoryObject): Directory. If None, the same
                directory is used
        Returns:
            (FileObject): file object of the new file
        """
        if new_file_name is None:
            if directory is None:
                raise NoDestinationError(
                    "Either new file name or directory must be specified"
                )
            new_file_name = self.file_name
        file_name, directory = self._resolve_directory_and_path(
            new_file_name, directory, default_directory=self.directory.path
        )
        new_file = FileObject(file_name, directory.path)
        shutil.copy(str(self.path), str(new_file.path))
        return new_file



from abc import ABCMeta


class HasPost(type):
    """
    A metaclass for adding a `__post__` method which has a compatible signature with
    `__init__` (and indeed receives all its input), but is guaranteed to be called
    only _after_ `__init__` is totally finished.

    Based on @jsbueno's reply in [this discussion](https://discuss.python.org/t/add-a-post-method-equivalent-to-the-new-method-but-called-after-init/5449/11)
    """

    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        if post := getattr(cls, "__post__", False):
            post(instance, *args, **kwargs)
        return instance


class AbstractHasPost(HasPost, ABCMeta):
    # Just for resolving metaclass conflic for ABC classes that have post
    pass
